@@ -1,8 +1,6 @@
 # -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:ft=tcl:et:sw=4:ts=4:sts=4
-# portfetch.tcl
-# $Id: portfetch.tcl 120931 2014-06-11 20:21:04Z ciserlohn@macports.org $
 #
-# Copyright (c) 2004 - 2012 The MacPorts Project
+# Copyright (c) 2004 - 2014, 2016-2018 The MacPorts Project
 # Copyright (c) 2002 - 2003 Apple Inc.
 # All rights reserved.
 #
@@ -48,7 +46,8 @@ namespace eval portfetch {
 }
 
 # define options: distname master_sites
-options master_sites patch_sites extract.suffix distfiles patchfiles use_bzip2 use_lzma use_xz use_zip use_7z use_lzip use_dmg dist_subdir \
+options master_sites patch_sites extract.suffix distfiles patchfiles use_tar \
+    use_bzip2 use_lzma use_xz use_zip use_7z use_lzip use_dmg dist_subdir \
     fetch.type fetch.user fetch.password fetch.use_epsv fetch.ignore_sslcert \
     master_sites.mirror_subdir patch_sites.mirror_subdir \
     bzr.url bzr.revision \
@@ -69,10 +68,10 @@ default fetch.type standard
 
 default bzr.cmd {[findBinary bzr $portutil::autoconf::bzr_path]}
 default bzr.dir {${workpath}}
-default bzr.revision {-1}
-default bzr.pre_args {"--builtin --no-aliases checkout --lightweight"}
+default bzr.revision -1
+default bzr.pre_args "--builtin --no-aliases checkout --lightweight --verbose"
 default bzr.args ""
-default bzr.post_args {"-r ${bzr.revision} ${bzr.url} ${worksrcdir}"}
+default bzr.post_args {-r ${bzr.revision} ${bzr.url} ${worksrcdir}}
 
 default cvs.cmd {[findBinary cvs $portutil::autoconf::cvs_path]}
 default cvs.password ""
@@ -82,47 +81,47 @@ default cvs.module {$distname}
 default cvs.tag ""
 default cvs.date ""
 default cvs.env {CVS_PASSFILE=${workpath}/.cvspass}
-default cvs.pre_args {"-z9 -f -d ${cvs.root}"}
+default cvs.pre_args {-z9 -f -d ${cvs.root}}
 default cvs.args ""
-default cvs.post_args {"${cvs.module}"}
+default cvs.post_args {${cvs.module}}
 
 default svn.cmd {[portfetch::find_svn_path]}
 default svn.dir {${workpath}}
 default svn.method {export}
 default svn.revision ""
 default svn.env {}
-default svn.pre_args {"--non-interactive --trust-server-cert"}
+default svn.pre_args "--non-interactive --trust-server-cert"
 default svn.args ""
 default svn.post_args ""
 
-default git.cmd {[findBinary git $portutil::autoconf::git_path]}
+default git.cmd {[portfetch::find_git_path]}
 default git.dir {${workpath}}
 default git.branch {}
 
 default hg.cmd {[findBinary hg $portutil::autoconf::hg_path]}
 default hg.dir {${workpath}}
-default hg.tag {tip}
+default hg.tag tip
 
 # Set distfiles
-default distfiles {[list [portfetch::suffix $distname]]}
+default distfiles {[portfetch::suffix $distname]}
 default dist_subdir {${name}}
 
 # user name & password
 default fetch.user ""
 default fetch.password ""
 # Use EPSV for FTP transfers
-default fetch.use_epsv "yes"
+default fetch.use_epsv yes
 # Ignore SSL certificate
-default fetch.ignore_sslcert "no"
+default fetch.ignore_sslcert no
 # Use remote timestamps
-default fetch.remote_time "no"
+default fetch.remote_time no
 
-default fallback_mirror_site "macports"
-default global_mirror_site "macports_distfiles"
-default mirror_sites.listfile {"mirror_sites.tcl"}
-default mirror_sites.listpath {"port1.0/fetch"}
+default global_mirror_site macports_distfiles
+default mirror_sites.listfile mirror_sites.tcl
+default mirror_sites.listpath port1.0/fetch
 
 # Option-executed procedures
+option_proc use_tar   portfetch::set_extract_type
 option_proc use_bzip2 portfetch::set_extract_type
 option_proc use_lzma  portfetch::set_extract_type
 option_proc use_xz    portfetch::set_extract_type
@@ -137,8 +136,14 @@ proc portfetch::set_extract_type {option action args} {
     global extract.suffix
     if {[string equal ${action} "set"] && [tbool args]} {
         switch $option {
+            use_tar {
+                set extract.suffix .tar
+            }
             use_bzip2 {
                 set extract.suffix .tar.bz2
+                if {![catch {findBinary lbzip2} result]} {
+                    depends_extract-append bin:lbzip2:lbzip2
+                }
             }
             use_lzma {
                 set extract.suffix .tar.lzma
@@ -181,14 +186,20 @@ proc portfetch::set_fetch_type {option action args} {
                 depends_fetch-append bin:cvs:cvs
             }
             svn {
-                if {${os.platform} == "darwin" && ${os.major} >= 10} {
+                # Sierra is the first macOS version whose svn supports modern TLS cipher suites.
+                if {${os.major} >= 16 || ${os.platform} ne "darwin"} {
                     depends_fetch-append bin:svn:subversion
                 } else {
                     depends_fetch-append port:subversion
                 }
             }
             git {
-                depends_fetch-append bin:git:git
+                # Mavericks is the first OS X version whose git supports modern TLS cipher suites.
+                if {${os.major} >= 13 || ${os.platform} ne "darwin"} {
+                    depends_fetch-append bin:git:git
+                } else {
+                    depends_fetch-append port:git
+                }
             }
             hg {
                 depends_fetch-append bin:hg:mercurial
@@ -199,14 +210,24 @@ proc portfetch::set_fetch_type {option action args} {
 
 proc portfetch::find_svn_path {args} {
     global prefix os.platform os.major
-    # Snow Leopard is the first Mac OS X version to include a recent enough svn (1.6.x) to support the --trust-server-cert option.
-    if {${os.platform} == "darwin" && ${os.major} >= 10} {
+    # Sierra is the first macOS version whose svn supports modern TLS cipher suites.
+    if {${os.major} >= 16 || ${os.platform} ne "darwin"} {
         return [findBinary svn $portutil::autoconf::svn_path]
     } else {
         return ${prefix}/bin/svn
     }
 }
 
+proc portfetch::find_git_path {args} {
+    global prefix os.platform os.major
+    # Mavericks is the first OS X version whose git supports modern TLS cipher suites.
+    if {${os.major} >= 13 || ${os.platform} ne "darwin"} {
+        return [findBinary git $portutil::autoconf::git_path]
+    } else {
+        return ${prefix}/bin/git
+    }
+}
+
 set_ui_prefix
 
 
@@ -272,12 +293,18 @@ proc portfetch::get_full_mirror_sites_path {} {
 # Perform the full checksites/checkpatchfiles/checkdistfiles sequence.
 # This method is used by distcheck target.
 proc portfetch::checkfiles {urls} {
-    global global_mirror_site fallback_mirror_site
+    global global_mirror_site ports_fetch_no-mirrors license
     upvar $urls fetch_urls
 
-    checksites [list patch_sites [list $global_mirror_site $fallback_mirror_site PATCH_SITE_LOCAL] \
-                master_sites [list $global_mirror_site $fallback_mirror_site MASTER_SITE_LOCAL]] \
-               [get_full_mirror_sites_path]
+    set sites [list patch_sites {} \
+                    master_sites {}]
+    if {(![info exists ports_fetch_no-mirrors] || ${ports_fetch_no-mirrors} eq "no") \
+            && [lsearch -exact -nocase $license "nomirror"] == -1} {
+        set sites [list patch_sites [list $global_mirror_site PATCH_SITE_LOCAL] \
+                        master_sites [list $global_mirror_site MASTER_SITE_LOCAL]]
+    }
+
+    checksites $sites [get_full_mirror_sites_path]
     checkpatchfiles fetch_urls
     checkdistfiles fetch_urls
 }
@@ -310,10 +337,10 @@ proc portfetch::bzrfetch {args} {
             return -code error [msgcat::mc "Bazaar checkout failed"]
         }
     } finally {
-        if ([info exists orig_http_proxy]) {
+        if {[info exists orig_http_proxy]} {
             set env(http_proxy) ${orig_http_proxy}
         }
-        if ([info exists orig_https_proxy]) {
+        if {[info exists orig_https_proxy]} {
             set env(HTTPS_PROXY) ${orig_https_proxy}
         }
     }
@@ -333,7 +360,7 @@ proc portfetch::cvsfetch {args} {
            patch_sites patchfiles filespath
 
     set cvs.args "${cvs.method} ${cvs.args}"
-    if {${cvs.method} == "export" && ![string length ${cvs.tag}] && ![string length ${cvs.date}]} {
+    if {${cvs.method} eq "export" && ![string length ${cvs.tag}] && ![string length ${cvs.date}]} {
         set cvs.tag "HEAD"
     }
     if {[string length ${cvs.tag}]} {
@@ -428,10 +455,10 @@ proc portfetch::gitfetch {args} {
     global worksrcpath patchfiles \
            git.url git.branch git.sha1 git.cmd
 
-    set options "-q"
-    if {[string length ${git.branch}] == 0} {
+    set options "--progress"
+    if {${git.branch} eq ""} {
         # if we're just using HEAD, we can make a shallow repo
-        set options "$options --depth=1"
+        append options " --depth=1"
     }
     set cmdstring "${git.cmd} clone $options ${git.url} ${worksrcpath} 2>&1"
     ui_debug "Executing: $cmdstring"
@@ -439,7 +466,7 @@ proc portfetch::gitfetch {args} {
         return -code error [msgcat::mc "Git clone failed"]
     }
 
-    if {[string length ${git.branch}] > 0} {
+    if {${git.branch} ne ""} {
         set env "GIT_DIR=${worksrcpath}/.git GIT_WORK_TREE=${worksrcpath}"
         set cmdstring "$env ${git.cmd} checkout -q ${git.branch} 2>&1"
         ui_debug "Executing $cmdstring"
@@ -483,7 +510,7 @@ proc portfetch::hgfetch {args} {
 proc portfetch::fetchfiles {args} {
     global distpath all_dist_files UI_PREFIX \
            fetch.user fetch.password fetch.use_epsv fetch.ignore_sslcert fetch.remote_time \
-           fallback_mirror_site portverbose usealtworkpath altprefix
+           portverbose
     variable fetch_urls
     variable urlmap
 
@@ -492,13 +519,13 @@ proc portfetch::fetchfiles {args} {
         lappend fetch_options -u
         lappend fetch_options "${fetch.user}:${fetch.password}"
     }
-    if {${fetch.use_epsv} != "yes"} {
+    if {${fetch.use_epsv} ne "yes"} {
         lappend fetch_options "--disable-epsv"
     }
-    if {${fetch.ignore_sslcert} != "no"} {
+    if {${fetch.ignore_sslcert} ne "no"} {
         lappend fetch_options "--ignore-ssl-cert"
     }
-    if {${fetch.remote_time} != "no"} {
+    if {${fetch.remote_time} ne "no"} {
         lappend fetch_options "--remote-time"
     }
     if {$portverbose eq "yes"} {
@@ -512,20 +539,12 @@ proc portfetch::fetchfiles {args} {
 
     foreach {url_var distfile} $fetch_urls {
         if {![file isfile "${distpath}/${distfile}"]} {
-            ui_info "$UI_PREFIX [format [msgcat::mc "%s doesn't seem to exist in %s"] $distfile $distpath]"
+            ui_info "$UI_PREFIX [format [msgcat::mc "%s does not exist in %s"] $distfile $distpath]"
             if {![file writable $distpath]} {
                 return -code error [format [msgcat::mc "%s must be writable"] $distpath]
             }
-            if {!$usealtworkpath && [file isfile ${altprefix}${distpath}/${distfile}]} {
-                if {[catch {file link -hard "${distpath}/${distfile}" "${altprefix}${distpath}/${distfile}"}]} {
-                    ui_debug "failed to hardlink ${distfile} into distpath, copying instead"
-                    file copy "${altprefix}${distpath}/${distfile}" "${distpath}/${distfile}"
-                }
-                ui_info "Found $distfile in ${altprefix}${distpath}"
-                continue
-            }
             if {!$sorted} {
-                sortsites fetch_urls [mirror_sites $fallback_mirror_site {} {} [get_full_mirror_sites_path]] master_sites
+                sortsites fetch_urls master_sites
                 set sorted yes
             }
             if {![info exists urlmap($url_var)]} {
@@ -533,20 +552,28 @@ proc portfetch::fetchfiles {args} {
                 set urlmap($url_var) $urlmap(master_sites)
             }
             unset -nocomplain fetched
+            set lastError ""
             foreach site $urlmap($url_var) {
                 ui_notice "$UI_PREFIX [format [msgcat::mc "Attempting to fetch %s from %s"] $distfile $site]"
                 set file_url [portfetch::assemble_url $site $distfile]
-                if {![catch {eval curl fetch $fetch_options {$file_url} {"${distpath}/${distfile}.TMP"}} result] &&
-                    ![catch {file rename -force "${distpath}/${distfile}.TMP" "${distpath}/${distfile}"} result]} {
+                try -pass_signal {
+                    curl fetch {*}$fetch_options $file_url "${distpath}/${distfile}.TMP"
+                    file rename -force "${distpath}/${distfile}.TMP" "${distpath}/${distfile}"
                     set fetched 1
                     break
-                } else {
-                    ui_debug "[msgcat::mc "Fetching distfile failed"]: $result"
+                } catch {{*} eCode eMessage} {
+                    ui_debug [msgcat::mc "Fetching distfile failed: %s" $eMessage]
+                    set lastError $eMessage
+                } finally {
                     file delete -force "${distpath}/${distfile}.TMP"
                 }
             }
             if {![info exists fetched]} {
-                return -code error [msgcat::mc "fetch failed"]
+                if {$lastError ne ""} {
+                    error $lastError
+                } else {
+                    error [msgcat::mc "fetch failed"]
+                }
             }
         }
     }
@@ -619,7 +646,7 @@ proc portfetch::fetch_main {args} {
     global all_dist_files fetch.type
 
     # Check for files, download if necessary
-    if {![info exists all_dist_files] && "${fetch.type}" == "standard"} {
+    if {![info exists all_dist_files] && "${fetch.type}" eq "standard"} {
         return 0
     }
 
